/*
 * showvideo.c - continuously displays jpeg images in an xwindow.
 * (for ov519-based webcams that return frames in jpeg format).
 * Copyright 2004 John McPherson (http://www.wlug.org.nz/JohnMcPherson)
 * Requires imlib and jpeg dev libraries, and v4l headers.
 * On debian sarge/sid: "apt-get install imlib11-dev libjpeg62-dev"
 * 
 * This program borrows code from 
 * # getjpeg.c (Copyright (c) 2003 Joerg Heckenbach <joerg@heckenbach-aw.de>),
 * # imlib primer
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <libgen.h>
// for jpeg - use raw jpeg library to decompress frame
#include <jpeglib.h>
// for imlib
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <Imlib.h>
// misc
#include <errno.h>

//#define USE_MMAP
#define DEF_WIDTH 640
#define DEF_HEIGHT 480

void usage (char *pname)
{
	fprintf (stderr,
	"Usage: %s <options>\n"
	" -d <device>          video device (default: /dev/video)\n"
/* works for some resolutions but not others... xxx todo fixme
  " -s NxN               define size of the output image (default:"
		" %dx%d)\n"
*/
//	"Example: %s -d /dev/video0 -s 640x480\n",
	"Example: %s -d /dev/video0\n",
	(char*)basename(pname), /*DEF_WIDTH, DEF_HEIGHT,*/
       	(char*)basename(pname));
	exit (1);
}



/* for jpeg library - handling buffered input for decompression */
static void init_source (j_decompress_ptr cinfo) { return; }
static boolean fill_input_buffer (j_decompress_ptr cinfo) { return 1; }
static void skip_input_data (j_decompress_ptr cinfo, long num_bytes) {
  if (num_bytes <= 0) return;
  cinfo->src->next_input_byte+=num_bytes;
  cinfo->src->bytes_in_buffer-=num_bytes;
}
static boolean resync_to_restart (j_decompress_ptr cinfo, int desired) { return 1; }
static void term_source (j_decompress_ptr cinfo) { return; }



int main(int argc, char **argv)
{

  int filedesc;
  //ssize_t count;
  int frame=-1;
  unsigned char *buffer;
  unsigned char *rgbdata;
  char *device = "/dev/video0";
  int c;
  int width = DEF_WIDTH, height = DEF_HEIGHT;
  /* for v4l/mmap reading */
  char *vmbuffer = NULL;
  unsigned short *vmpointer = NULL;
  struct video_mbuf vmbuf;
  struct video_mmap vmmap;
  
  /* for jpeg */
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  struct jpeg_source_mgr source;

  /* for imlib */
  Display *disp;
  ImlibData *im_id;
  XSetWindowAttributes attr;
  Window win;
  ImlibImage *image;
  

  while ((c = getopt (argc, argv, "d:s:n:")) != EOF) {
    switch (c) {
    case 'd': /* change default device */
      device = optarg;
      break;
    case 's':
      sscanf (optarg, "%dx%d", &width, &height);
      break;
    default:
      usage (argv[0]);
      break;
    }
  }

  /* initialise X window and imlib stuff */
  /* Connect to the default Xserver */
  disp=XOpenDisplay(NULL);
  /* Immediately afterwards Initialise Imlib */
  im_id=Imlib_init(disp);
  /* Create a Window to display in */
  win=XCreateWindow(disp,DefaultRootWindow(disp),0,0,width,height,
		    0,im_id->x.depth,
		    InputOutput,im_id->x.visual,0,&attr);
  /* Actually display the window */
  XMapWindow(disp,win);
  /* Synchronise with the Xserver */
  XSync(disp,False);
  
  /* init jpeg library */
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);
  cinfo.src=&source;
  /* functions for managing jpeg lib's input buffer */
  source.init_source=init_source;
  source.fill_input_buffer=fill_input_buffer;
  source.skip_input_data=skip_input_data;
  source.resync_to_restart=resync_to_restart;
  source.term_source=term_source;
  
  rgbdata=malloc(height * width * 3);
  // uninitialised memory for now...
  image=Imlib_create_image_from_data(im_id, rgbdata, NULL /*alpha*/, width, height);
  filedesc=open(device, O_RDWR);
  if (filedesc < 0) {
    printf("open failed: %s\n", strerror(errno));
    exit(1);
  }
  
  ioctl(filedesc, VIDIOCGMBUF, &vmbuf);
  vmbuffer = mmap(0, vmbuf.size, PROT_READ, MAP_SHARED, filedesc, 0);
  if (!(vmbuffer))
    perror("mmap failed");

  vmmap.format = VIDEO_PALETTE_RGB24;
  vmmap.frame = 0;
  vmmap.width = width;
  vmmap.height = height;

  if (ioctl (filedesc, VIDIOCMCAPTURE, &vmmap) == -1) {
    perror ("VIDIOCMCAPTURE(0)");
    // this happens to me if I try to run this program on my tv tuner card
    printf("device %s isn't a webcam?\n", device);
    exit(1);
  }
  vmmap.frame = 1;

  while (1) {
    frame++;

    if (ioctl (filedesc, VIDIOCMCAPTURE, &vmmap) == -1) {
      perror ("VIDIOCMCAPTURE()");
    }
    vmmap.frame = (vmmap.frame + 1) % vmbuf.frames;
    if (ioctl (filedesc, VIDIOCSYNC, &vmmap.frame) == -1) {
      perror ("VIDIOCSYNC()");
    }

    // The length / 8 of the JPEG is stored in the first 2 bytes of vmbuffer
    vmpointer = (unsigned short *)(vmbuffer+vmbuf.offsets[vmmap.frame]);

    source.next_input_byte=vmbuffer+vmbuf.offsets[vmmap.frame] + 2;
    source.bytes_in_buffer=(ssize_t)((unsigned int)(vmpointer[0])<<3); // count

#if 0
    if (fwrite(vmbuffer+vmbuf.offsets[vmmap.frame] + 2,
	       sizeof(unsigned char),
	       (size_t)count, ofid) == count) {
#endif



      // read in a jpeg frame from the interface

      jpeg_read_header(&cinfo, TRUE); // will use previously buffered input?
      jpeg_start_decompress(&cinfo);

      rgbdata=image->rgb_data;
      while (cinfo.output_scanline < cinfo.output_height) {
	    // point to next row of image data
	    buffer=rgbdata + (cinfo.output_scanline*width*3);
	    jpeg_read_scanlines(&cinfo, &buffer, height-cinfo.output_scanline);
	  }

	  //printf("%dx%d\n", width, height);
	  jpeg_finish_decompress(&cinfo);

	  if (!image) {
	    printf ("Error creating image from data\n");
	    continue;
	  }
	  // update the image since we changed the raw char* data
	  Imlib_apply_modifiers_to_rgb(im_id,image);
	  /* Put the Image in the window, at the window's size and apply a shape mask */
	  Imlib_apply_image(im_id,image,win);
	  //XSync(disp,True);
	}
	jpeg_destroy_decompress(&cinfo);
	close(filedesc);
	sleep(1);

	return 0;
}
